from __future__ import annotations

import typing
from typing import Any, Generic, Iterable, List, Mapping

from di.api.dependencies import DependantBase, DependencyParameter
from di.api.providers import DependencyType

Dependency = Any


class SolvedDependant(Generic[DependencyType]):
    """Representation of a fully solved dependency as DAG"""

    __slots__ = ("dependency", "dag", "topsort", "container_cache")

    def __init__(
        self,
        dependency: DependantBase[DependencyType],
        dag: Mapping[DependantBase[Any], List[DependencyParameter]],
        # A toplogical sort represented as groups of dependencies that can be executed concurrently
        # This is just one of many possible valid toplogical sorts
        # It is calculated optimistically: this is how dependencies _would_ be executed
        # if all dependencies executed instantly
        topsort: Iterable[Iterable[DependantBase[Any]]],
        # container_cache can be used by the creating container to store data that is tied
        # to the SolvedDependant
        container_cache: typing.Any,
    ):
        self.dependency = dependency
        self.dag = dag
        self.topsort = topsort
        self.container_cache = container_cache

    def get_flat_subdependants(self) -> List[DependantBase[Any]]:
        """Get an exhaustive list of all of the dependencies of this dependency,
        in no particular order.
        """
        return list(self.dag.keys() - {self.dependency})

    def __hash__(self) -> int:
        return id(self)

    def __eq__(self, __o: object) -> bool:
        return __o is self



from __future__ import annotations

import threading
from contextlib import AsyncExitStack, ExitStack, asynccontextmanager, contextmanager
from typing import (
    Any,
    Deque,
    Dict,
    Generator,
    Iterable,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

from graphlib2 import TopologicalSorter

from di._utils.inspect import is_async_gen_callable, is_gen_callable
from di.api.dependencies import DependantBase
from di.api.executor import AsyncTask as ExecutorAsyncTask
from di.api.executor import SyncTask as ExecutorSyncTask
from di.api.executor import Task as ExecutorTask
from di.api.providers import DependencyProvider
from di.api.scopes import Scope
from di.exceptions import IncompatibleDependencyError


class ExecutionState:
    __slots__ = (
        "stacks",
        "results",
        "toplogical_sorter",
        "lock",
    )

    def __init__(
        self,
        stacks: Mapping[Scope, Union[AsyncExitStack, ExitStack]],
        results: Dict[Union[AsyncTask, SyncTask], Any],
        toplogical_sorter: TopologicalSorter[Union[AsyncTask, SyncTask]],
    ):
        self.stacks = stacks
        self.results = results
        self.toplogical_sorter = toplogical_sorter
        self.lock = threading.Lock()


DependencyType = TypeVar("DependencyType")

TaskQueue = Deque[Optional[ExecutorTask]]


def gather_new_tasks(
    state: ExecutionState,
) -> Generator[Optional[ExecutorTask], None, None]:
    """Look amongst our dependants to see if any of them are now dependency free"""
    res = state.results
    ts = state.toplogical_sorter
    while True:
        if not ts.is_active():
            yield None
            return
        ready = ts.get_ready()
        if not ready:
            break
        marked = False
        for t in ready:
            if t in res:
                # task was passed in by value or cached
                ts.done(t)
                marked = True
            else:
                yield t
        if not marked:
            # we didn't mark any nodes as done
            # so there's no point in calling get_ready() again
            break
    if not ts.is_active():
        yield None


class Task:
    __slots__ = (
        "dependant",
        "call",
        "positional_parameters",
        "keyword_parameters",
        "scope",
    )
    call: DependencyProvider
    scope: Scope

    def __init__(
        self,
        dependant: DependantBase[Any],
        positional_parameters: Iterable[Union[AsyncTask, SyncTask]],
        keyword_parameters: Iterable[Tuple[str, Union[AsyncTask, SyncTask]]],
    ) -> None:
        self.dependant = dependant
        self.scope = self.dependant.scope
        assert dependant.call is not None
        self.call = dependant.call
        self.positional_parameters = positional_parameters
        self.keyword_parameters = keyword_parameters

    def gather_params(
        self, results: Dict[Union[AsyncTask, SyncTask], Any]
    ) -> Tuple[Iterable[Any], Mapping[str, Any]]:
        """Gather all parameters (aka *args and **kwargs) needed for computation"""
        return (
            (results[t] for t in self.positional_parameters),
            {k: results[t] for k, t in self.keyword_parameters},
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(self.dependant)})"


class AsyncTask(Task, ExecutorAsyncTask):
    __slots__ = ("is_generator",)

    def __init__(
        self,
        dependant: DependantBase[Any],
        positional_parameters: Iterable[Union[AsyncTask, SyncTask]],
        keyword_parameters: Iterable[Tuple[str, Union[AsyncTask, SyncTask]]],
    ) -> None:
        super().__init__(dependant, positional_parameters, keyword_parameters)
        self.is_generator = is_async_gen_callable(self.call)
        if self.is_generator:
            self.call = asynccontextmanager(self.call)  # type: ignore[arg-type]

    async def compute(  # type: ignore[override]  # we do it this way to avoid exposing implementation details to users
        self,
        state: ExecutionState,
    ) -> Iterable[Optional[ExecutorTask]]:
        args, kwargs = self.gather_params(state.results)

        if self.is_generator:
            try:
                enter = state.stacks[self.scope].enter_async_context  # type: ignore[union-attr]
            except AttributeError:
                raise IncompatibleDependencyError(
                    f"The dependency {self.dependant} is an awaitable dependency"
                    f" and canot be used in the sync scope {self.dependant.scope}"
                )
            state.results[self] = await enter(
                self.call(*args, **kwargs)  # type: ignore[arg-type]
            )
        else:
            state.results[self] = await self.call(*args, **kwargs)  # type: ignore[misc]
        state.toplogical_sorter.done(self)
        return gather_new_tasks(state)


class SyncTask(Task, ExecutorSyncTask):
    __slots__ = ("is_generator",)

    def __init__(
        self,
        dependant: DependantBase[Any],
        positional_parameters: Iterable[Union[AsyncTask, SyncTask]],
        keyword_parameters: Iterable[Tuple[str, Union[AsyncTask, SyncTask]]],
    ) -> None:
        super().__init__(dependant, positional_parameters, keyword_parameters)
        self.is_generator = is_gen_callable(self.call)
        if self.is_generator:
            self.call = contextmanager(self.call)  # type: ignore[arg-type]

    def compute(  # type: ignore[override]  # we do it this way to avoid exposing implementation details to users
        self,
        state: ExecutionState,
    ) -> Iterable[Optional[ExecutorTask]]:
        args, kwargs = self.gather_params(state.results)

        if self.is_generator:
            state.results[self] = state.stacks[self.scope].enter_context(
                self.call(*args, **kwargs)  # type: ignore[arg-type]
            )
        else:
            state.results[self] = self.call(*args, **kwargs)
        state.toplogical_sorter.done(self)
        return gather_new_tasks(state)



import sys
from typing import Any, Collection, ContextManager, Mapping, Optional, TypeVar

if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol

from di._utils.types import FusedContextManager
from di.api.dependencies import DependantBase, DependencyType
from di.api.providers import DependencyProvider, DependencyProviderType
from di.api.scopes import Scope
from di.api.solved import SolvedDependant

ContainerType = TypeVar("ContainerType")


class ContainerProtocol(Protocol):
    @property
    def scopes(self) -> Collection[Scope]:
        ...

    def bind(
        self,
        provider: DependantBase[DependencyType],
        dependency: DependencyProviderType[DependencyType],
    ) -> ContextManager[None]:
        """Bind a new dependency provider for a given dependency.

        This can be used as a function (for a permanent bind, cleared when `scope` is exited)
        or as a context manager (the bind will be cleared when the context manager exits).

        Binds are only identified by the identity of the callable and do not take into account
        the scope or any other data from the dependency they are replacing.
        """
        ...

    def solve(
        self,
        dependency: DependantBase[DependencyType],
    ) -> SolvedDependant[DependencyType]:
        """Solve a dependency.

        Returns a SolvedDependant that can be executed to get the dependency's value.
        """
        ...

    def enter_scope(
        self: ContainerType, scope: Scope
    ) -> FusedContextManager[ContainerType]:
        """Enter a scope and get back a new BaseContainer in that scope"""
        ...

    def execute_sync(
        self,
        solved: SolvedDependant[DependencyType],
        *,
        values: Optional[Mapping[DependencyProvider, Any]] = None,
    ) -> DependencyType:
        """Execute an already solved dependency.

        This method is synchronous and uses a synchronous executor,
        but the executor may still be able to execute async dependencies.
        """
        ...

    async def execute_async(
        self,
        solved: SolvedDependant[DependencyType],
        *,
        values: Optional[Mapping[DependencyProvider, Any]] = None,
    ) -> DependencyType:
        """Execute an already solved dependency."""
        ...



"""Convenience functions, mainly for the purpose of providing proper type annotations for default arguments.
"""

from typing import Optional, overload

from di.api.providers import (
    AsyncGeneratorProvider,
    CallableProvider,
    CoroutineProvider,
    DependencyProviderType,
    DependencyType,
    GeneratorProvider,
)
from di.api.scopes import Scope
from di.dependant import Dependant


@overload
def Depends(
    call: Optional[AsyncGeneratorProvider[DependencyType]] = None,
    *,
    scope: Scope = None,
    share: bool = True,
    wire: bool = True,
    autowire: bool = True,
    sync_to_thread: bool = False,
) -> DependencyType:
    ...


@overload
def Depends(
    call: Optional[CoroutineProvider[DependencyType]] = None,
    *,
    scope: Scope = None,
    share: bool = True,
    wire: bool = True,
    autowire: bool = True,
    sync_to_thread: bool = False,
) -> DependencyType:
    ...


@overload
def Depends(
    call: Optional[GeneratorProvider[DependencyType]] = None,
    *,
    scope: Scope = None,
    share: bool = True,
    wire: bool = True,
    autowire: bool = True,
    sync_to_thread: bool = False,
) -> DependencyType:
    ...


@overload
def Depends(
    call: Optional[CallableProvider[DependencyType]] = None,
    *,
    scope: Scope = None,
    share: bool = True,
    wire: bool = True,
    autowire: bool = True,
    sync_to_thread: bool = False,
) -> DependencyType:
    ...


def Depends(
    call: Optional[DependencyProviderType[DependencyType]] = None,
    *,
    scope: Scope = None,
    share: bool = True,
    wire: bool = True,
    autowire: bool = True,
    sync_to_thread: bool = False,
) -> DependencyType:
    return Dependant(  # type: ignore[return-value]
        call=call,
        scope=scope,
        share=share,
        wire=wire,
        autowire=autowire,
        sync_to_thread=sync_to_thread,
    )



from typing import Dict, Iterable, List, TypeVar

from graphlib2 import TopologicalSorter

T = TypeVar("T")


def topsort(graph: Dict[T, List[T]]) -> Iterable[Iterable[T]]:
    ts = TopologicalSorter(graph)
    ts.prepare()
    result: List[Iterable[T]] = []
    while ts.is_active():
        ready = ts.get_ready()
        result.append(ready)
        ts.done(*ready)
    return result
